# """
# 3D camera motion
#
# These functions allow Ren'Py to simulate 3D camera motion.  If you use this
# feature, in the first, set additional layers.
#
#    init -1 python:
#        config.layers = [ 'master', 'l1', 'l2', 'l3', 'transient', 'screens', 'overlay']
#
# Second, register layers which participate to a 3D motion as 3d layers.
#    init python:
#        register_3d_layer( ('master', 1000), ('l1', 300), ('l2', 500), ('l3', 700))
#
# Third, move a camera
#    label start:
#        scene bg
#        show A onlayer l1
#        show B onlayer l2
#        show B onlayer l3
#        'Start motion'
#        'It takes 1 second to move a camera to (0, 130, 0)'
#        $ camera_move(0, 130, 0, 0, 1)
#        'It takes 5 seconds to move a camera to (-70, 130, 1000)'
#        $ camera_move(-70, 130, 1000, 0, 5)
#        'A camera moves to (0, 0, 0) at the moment'
#        $ camera_move(0, 0, 0)
#        'It takes 1 second to rotate a camera to 360 degrees'
#        $ camera_move(0, 0, 0, 360, 1)
#        'reset motion'
#        $ camera_reset()
#        'end'
#
# You also use GUI viewer. Press shift+c and open Camera Viewer if config.developer is True.
# Transform Viewer also helps you.
#
# Notice that 3D camera motion and transforms to a layer can't be usable at the same time.
#
# This is beta version, rotate doesn't work yes. A camera doesn't move exactly when camera is too close to a subject.
# """

init -1600 python:

    # value of _3d_layers isn't included in rollback.
    # so I copy it many times.
    _3d_layers = {}
    _3d_layers_org = {}
    _camera_x = 0
    _camera_y = 0
    _camera_z = 0
    _camera_r = 0
    _film_size = 50


    def register_3d_layer(*tuples):
        """
         :doc: camera

         Register layers as 3D layers and those distances from a camera. 3D
         layers is applied transforms to by a camera motion. If you want to
         change distances in game, register again.

         `tuples`
              These should be a tuple or tuples composed of a layer name and
              the distance from a camera.
         """
        global _3d_layers
        _3d_layers = {}
        for t in tuples:
            # [xanchor, yanchor, zoom, rotate, z]
            if t[1] == 0:
                t[1] = 1
            _3d_layers[t[0]] = [.5, .5, 1., 0, float(t[1])]

    def release_3d_layer(layer):
        """
         :doc: camera

         Release 3D layers and the transforms of it.
         """
        global _3d_layers
        # renpy.show_layer_at(Transform(xalign=.5, yalign=.5, zoom=1.0, rotate=0, transform_anchor=True), layer)
        renpy.show_layer_at(Transform(xalign=.5, yalign=.5, zoom=1.0), layer)
        del _3d_layers[layer]

    # reset a cumulative error
    def camera_reset():
        """
         :doc: camera

         Reset the transforms of all 3D layers.
         """
        global _3d_layers, _camera_x, _camera_y, _camera_z, _camera_r
        _camera_x = 0
        _camera_y = 0
        _camera_z = 0
        _camera_r = 0
        af_3d_layers = {}
        for l in sorted(_3d_layers.keys()):
            af_3d_layers[l] = [.5, .5, 1., 0, _3d_layers[l][4]]
            # renpy.show_layer_at(Transform(xalign=.5, yalign=.5, zoom=1.0, rotate=0, transform_anchor=True), l)
            renpy.show_layer_at(Transform(xalign=.5, yalign=.5, zoom=1.0), l)
        _3d_layers = af_3d_layers

    def set_film_size(film_size):
        """
         :doc: camera

         Set the film_size of camera.

         `film_size`
              This is used as the film_size of a camera.
         """
        global _film_size
        _film_size = film_size
        camera_reset()

    def camera_move(x, y, z, rotate=0, during=0, warper='linear'):
        """
         :doc: camera

         Move the coordinate of a camera and apply transforms to all 3D layers.

         `x`
              the x coordinate of a camera
         `y`
              the y coordinate of a camera
         `z`
              the z coordinate of a camera
         `rotate`
              Defaul 0, the rotate of a camera
         `during`
              Default 0, this is the second times taken to move a camera.
         `warper`
              Default 'linear', this should be string and the name of a warper
              registered with ATL.
         """
        global _3d_layers, _camera_x, _camera_y, _camera_z, _camera_r, _film_size

        # adjustment x, y
        x = x/100.
        y = y/100.
        x0 = _camera_x/100.
        y0 = _camera_y/100.
        z0 = _camera_z

        af_3d_layers = {}
        for l in sorted(_3d_layers.keys()):
            ax0 = _3d_layers[l][0]
            ay0 = _3d_layers[l][1]
            zoom0 = _3d_layers[l][2]
            r0 = _3d_layers[l][3]
            l_z = _3d_layers[l][4]

            distance0 = float(_camera_z + l_z)
            distance = float(z + l_z)

            # time where a camera crosses subject
            cross_time = -1.
            if (distance < 0 and distance0 > 0) or (distance > 0 and distance0 < 0):
                cross_time = -distance0/(distance - distance0)
            elif distance < 0 and distance0 < 0:
                cross_time = 0

            # avoid zero division
            if distance == 0:
                distance = 0.1

            ax = ( _film_size*(x-x0) + distance0*(ax0-0.5) )/distance + 0.5
            ay = ( _film_size*(y-y0) + distance0*(ay0-0.5) )/distance + 0.5
            zoom  = (distance0/distance)*zoom0

            if during > 0:
                renpy.show_layer_at(Transform(function=renpy.curry(_camera_trans)(ax0=ax0, ax=ax, ay0=ay0, ay=ay, distance0=distance0, x0=x0, x=x, y0=y0, y=y, z0=z0, z=z, zoom0=zoom0, zoom=zoom, r0=r0, r=rotate, during=during, warper=warper, cross_time=cross_time)), l)
            else:
                if zoom >= 0:
                    # renpy.show_layer_at(Transform(xpos=.5, ypos=.5, xanchor=ax, yanchor=ay, zoom=zoom, rotate=r, transform_anchor=True, alpha=1), l)
                    renpy.show_layer_at(Transform(xpos=.5, ypos=.5, xanchor=ax, yanchor=ay, zoom=zoom, rotate=rotate, alpha=1), l)
                else:
                    # renpy.show_layer_at(Transform(xpos=.5, ypos=.5, xanchor=ax, yanchor=ay, zoom=zoom, rotate=r, transform_anchor=True, alpha=0), l)
                    renpy.show_layer_at(Transform(xpos=.5, ypos=.5, xanchor=ax, yanchor=ay, zoom=zoom, rotate=rotate, alpha=0), l)

            af_3d_layers[l] = [ax, ay, zoom, rotate, l_z]

        _camera_x = int(x*100)
        _camera_y = int(y*100)
        _camera_z = z
        _camera_r = rotate
        _3d_layers = af_3d_layers

    def _camera_trans(tran, st, at, ax0, ax, ay0, ay, distance0, x0, x, y0, y, z0, z, zoom0, zoom, r0, r, during, warper, cross_time):
        g = renpy.atl.warpers[warper](st/during) 
        if g <= 1:
            distance = g*(z-z0)+distance0
            if distance == 0:
                distance = .1
            tran.xpos    = .5 
            tran.ypos    = .5

            tran.xanchor = ( _film_size*g*(x-x0) + distance0*(ax0-0.5) )/distance + 0.5
            tran.yanchor = ( _film_size*g*(y-y0) + distance0*(ay0-0.5) )/distance + 0.5
            # tran.xanchor = _film_size*((g*(x-x0)+x0)/distance - x0/distance0) + ax0
            # tran.yanchor = _film_size*((g*(y-y0)+y0)/distance - y0/distance0) + ay0

            # tran.transform_anchor = True
            tran.rotate  = g*(r  -  r0) + r0

            if zoom0 > 0 and zoom < 0:
                if cross_time <= g:
                    tran.alpha = 0
                else:
                    tran.alpha = 1
                    tran.zoom    = (distance0 / distance)*zoom0
            elif zoom0 < 0 and zoom > 0:
                if cross_time <= g:
                    tran.alpha = 1
                    tran.zoom    = (distance0 / distance)*zoom0
                else:
                    tran.alpha = 0
            elif zoom0 < 0:
                tran.alpha = 0
            elif zoom0 > 0:
                tran.alpha = 1
                tran.zoom    = (distance0 / distance)*zoom0
            return .005
        else:
            # tran.transform_anchor = True
            tran.xpos    = .5 
            tran.ypos    = .5
            tran.xanchor = ax
            tran.yanchor = ay
            tran.zoom    = zoom
            tran.rotate    = r
            if zoom >= 0:
                tran.alpha = 1
            else:
                tran.alpha = 0
            return None


    class CameraViewer(object):
        def __init__(self, range_x=5000, range_y=5000, range_z=10000, range_r=360, range_film_size=100, range_layer_z=10000):

            self.range_x         = range_x
            self.range_y         = range_y
            self.range_z         = range_z
            self.range_r         = range_r
            self.range_film_size = range_film_size
            self.range_layer_z   = range_layer_z

        def viewer(self):
            global _3d_layers, _3d_layers_org, _film_size, _film_size_org
            if not config.developer:
                return
            _3d_layers_org = _3d_layers.copy()
            _film_size_org = _film_size
            camera_reset()
            renpy.call_in_new_context("_camera_viewer", viewer=self)

        def reset(self):
            global _3d_layers, _3d_layers_org, _film_size
            for l in _3d_layers:
                _3d_layers[l][0] = _3d_layers_org[l][0]
                _3d_layers[l][1] = _3d_layers_org[l][1]
                _3d_layers[l][2] = _3d_layers_org[l][2]
                _3d_layers[l][3] = _3d_layers_org[l][3]
            _film_size = _film_size_org
            camera_reset()

        def x_changed(self, v):
            camera_move(v - self.range_x, _camera_y, _camera_z, _camera_r)
            renpy.restart_interaction()

        def y_changed(self, v):
            camera_move(_camera_x, v - self.range_y, _camera_z, _camera_r)
            renpy.restart_interaction()

        def z_changed(self, v):
            camera_move(_camera_x, _camera_y, v - self.range_z, _camera_r)
            renpy.restart_interaction()

        def r_changed(self, v):
            camera_move(_camera_x, _camera_y, _camera_z, v - self.range_r)
            renpy.restart_interaction()

        def f_changed(self, v):
            global _film_size
            _film_size = v
            camera_reset()
            renpy.restart_interaction()
            
        def generate_layer_z_changed(self, l):
            def layer_z_changed(v):
                if v == 0:
                    v = 1
                _3d_layers[l][4] = float(v)
                camera_reset()
                renpy.restart_interaction()
            return layer_z_changed

label _camera_viewer(viewer):
    show screen _camera_viewer(viewer)
    $ui.interact()

    return

screen _camera_viewer(viewer):
    key "game_menu" action Return()
    zorder 10

    frame:
        yalign .9
        background "#0002"
        has vbox
        text "Camera Setting"
        hbox:
            xfill True
            text "x     {:>7}".format(_camera_x)
            null width 10
            bar adjustment ui.adjustment(range=viewer.range_x*2, value=_camera_x+viewer.range_x, page=1, changed=viewer.x_changed) xalign 1.
        hbox:
            xfill True
            text "y     {:>7}".format(_camera_y)
            null width 10
            bar adjustment ui.adjustment(range=viewer.range_y*2, value=_camera_y+viewer.range_y, page=1, changed=viewer.y_changed) xalign 1.
        hbox:
            xfill True
            text "z     {:>7}".format(_camera_z)
            null width 10
            bar adjustment ui.adjustment(range=viewer.range_z*2, value=_camera_z+viewer.range_z, page=1, changed=viewer.z_changed) xalign 1.
        hbox:
            xfill True
            text "rotate{:>7}".format(_camera_r)
            null width 10
            bar adjustment ui.adjustment(range=viewer.range_r*2, value=_camera_r+viewer.range_r, page=1, changed=viewer.r_changed) xalign 1.
        hbox:
            xfill True
            text "film size {:>3}".format(_film_size)
            null width 10
            bar adjustment ui.adjustment(range=viewer.range_film_size, value=_film_size, page=1, changed=viewer.f_changed) xalign 1.
        null height 10
        text _("3D layers z")
        for l in sorted(_3d_layers.keys()):
            hbox:
                xfill True
                text "{:<6}{:>7d}".format(l, int(_3d_layers[l][4]))
                null width 10
                bar adjustment ui.adjustment(range=viewer.range_layer_z, value=_3d_layers[l][4], page=1, changed=viewer.generate_layer_z_changed(l)) xalign 1.
        hbox:
            xalign 1.
            textbutton _("reset") action [camera_reset, renpy.restart_interaction]
            textbutton _("close") action Return()

#"""
# Transform Viewer
# 
# Press shift+t and open Transform Viewer if config.developer is True.
# This allow you to adjustment transform properties by bar and see the result at
# the moment.
# 
# Notice that a 'reset' button doesn't work always.
#"""

init -1600 python:
    class TransformViewer(object):
        def __init__(self):

            self.int_range = 1500
            self.float_range = 7.0
            # layer->tag->property->value
            self.state_org = {}
            # {property:default, float}, default is used when property can't be got.
            self.props_list = (
            "xpos",
            "ypos",
            "xanchor",
            "yanchor",
            "xzoom",
            "yzoom",
            "zoom",
            "rotate",
            "alpha",
            "additive"
            )
            self.props_default = {
            "xpos":(0., False),
            "ypos":(0., False),
            "xanchor":(0., False),
            "yanchor":(0., False),
            "xzoom":(1., True),
            "yzoom":(1., True),
            "zoom":(1., True),
            "rotate":(0, False),
            "alpha":(1., True),
            "additive":(0., True),
            }

        def viewer(self):
            if not config.developer:
                return
            sle = renpy.game.context().scene_lists
            # back up for reset()
            for layer in config.layers:
                self.state_org[layer] = {}
                for tag in sle.layers[layer]:
                    d = sle.get_displayable_by_tag(layer, tag[0])
                    pos = renpy.get_placement(d)
                    state = getattr(d, "state", None)

                    self.state_org[layer][tag[0]] = {}
                    for p in ["xpos", "ypos", "xanchor", "yanchor"]:
                        self.state_org[layer][tag[0]][p] = getattr(pos, p, None)
                    for p in self.props_list:
                        if not self.state_org[layer][tag[0]].has_key(p):
                            self.state_org[layer][tag[0]][p] = getattr(state, p, None)
                        if self.state_org[layer][tag[0]][p] is None:
                            self.state_org[layer][tag[0]][p] = self.props_default[p][0]
            renpy.call_in_new_context("_transform_viewer", viewer=self)

        def reset(self):
            for layer in config.layers:
                if layer in ['transient', 'screens', 'overlay']:
                    break
                for tag, props in self.state_org[layer].iteritems():
                    if tag and props:
                        if props["rotate"] == 0:
                            props["rotate"] = None
                        renpy.show(tag, [Transform(**props)], layer=layer)
            renpy.restart_interaction()

        def generate_changed(self, tag, prop, layer, int=False):
            def changed(v):
                kwargs = {}
                d = renpy.game.context().scene_lists.get_displayable_by_tag(layer, tag)
                pos = renpy.get_placement(d)
                state = getattr(d, "state", None)

                for p in self.props_list:
                    kwargs[p] = getattr(pos, p, None)
                    if kwargs[p] is None:
                        kwargs[p] = getattr(state, p, None)

                if int and not self.props_default[prop][1]:
                    kwargs[prop] = v - self.int_range
                else:
                    kwargs[prop] = v -self.float_range

                for p in self.props_list:
                    if kwargs[p] is None:
                        del kwargs[p]
                renpy.show(tag, [Transform(**kwargs)], layer=layer)
                renpy.restart_interaction()
            return changed

        def get_state(self, layer, tag, p):
            sle = renpy.game.context().scene_lists

            if tag:
                d = sle.get_displayable_by_tag(layer, tag)
                pos = renpy.get_placement(sle.get_displayable_by_tag(layer, tag))
                state = getattr(pos, p, None)
                if state is None:
                    state = getattr(getattr(d, "state", None), p, None)
                # set default
                if state is None:
                    state = self.props_default[p][0]
                if self.props_default[p][1]:
                    state = float(state)
                return state
            return None

label _transform_viewer(viewer):
    show screen _transform_viewer(viewer)
    $ui.interact()

    return

screen _transform_viewer(viewer, layer="master", tag=""):
    key "game_menu" action Return()
    zorder 10

    frame:
        xfill True
        ysize 500
        background "#0002"
        vbox:

            hbox:
                text _("layers")
                for l in config.layers:
                    if l not in ["screens", "transient", 'overlay', "debug"]:
                        textbutton "[l]" action [SelectedIf(l == layer), Show("_transform_viewer", viewer=viewer, layer=l)]

            hbox:
                text _("images")
                for t in renpy.game.context().scene_lists.layers[layer]:
                    textbutton "[t[0]]" action [SelectedIf(t[0] == tag), Show("_transform_viewer", viewer=viewer, layer=layer, tag=t[0])]

            if tag:
                # text "[tag]"
                for p in viewer.props_list:
                    if isinstance(viewer.get_state(layer, tag, p), int):
                        hbox:
                            xfill True
                            text "[p:<8]{:>10}".format(viewer.get_state(layer, tag, p))
                            null width 10
                            bar adjustment ui.adjustment(range=viewer.int_range*2, value=viewer.get_state(layer, tag, p)+viewer.int_range, page=1, changed=viewer.generate_changed(tag, p, layer, True)) xalign 1.
                    elif isinstance(viewer.get_state(layer, tag, p), float):
                        hbox:
                            xfill True
                            text "[p:<8]{:>10.4f}".format(viewer.get_state(layer, tag, p))
                            null width 10
                            bar adjustment ui.adjustment(range=viewer.float_range*2, value=viewer.get_state(layer, tag, p)+viewer.float_range, page=.05, changed=viewer.generate_changed(tag, p, layer)) xalign 1.
        hbox:
            xalign 1.
            yalign 1.
            textbutton _("reset") action [viewer.reset, renpy.restart_interaction]
            textbutton _("close") action Return()

init -1600 python:
    def _scene(layers=['master']):

        # Also clear 3d layers
        if not isinstance(layers, list):
            layers = [layers]
        renpy.scene(layers + list(_3d_layers))
    config.scene = _scene
    _camera_viewer = CameraViewer()
    _transform_viewer = TransformViewer()

init -1098 python:
    # overwrite keymap
    km = renpy.Keymap(
        rollback = renpy.rollback,
        screenshot = _screenshot,
        toggle_fullscreen = renpy.toggle_fullscreen,
        toggle_skip = _keymap_toggle_skipping,
        fast_skip = _fast_skip,
        game_menu = _invoke_game_menu,
        hide_windows = renpy.curried_call_in_new_context("_hide_windows"),
        launch_editor = _launch_editor,
        reload_game = _reload_game,
        developer = _developer,
        quit = renpy.quit_event,
        iconify = renpy.iconify,
        help = _help,
        choose_renderer = renpy.curried_call_in_new_context("_choose_renderer"),
        console = _console.enter,
        profile_once = _profile_once,
        self_voicing = Preference("self voicing", "toggle"),
        # variable_viewer = _dynamic_watch,
        camera_viewer = CameraViewer().viewer,
        transform_viewer = TransformViewer().viewer
        )

    config.underlay = [ km ]

    del km

init 1000 python:
    config.locked = False
    config.keymap["camera_viewer"] = ['C']
    config.keymap["transform_viewer"] = ['T']
    config.locked = True

    def scene(layer='master'):

        if renpy.game.context().init_phase:
            raise Exception("Scene may not run while in init phase.")

        sls = renpy.display.core.scene_lists()
        # Also clear layers registered as 3d layer
        for layer in [layer] + list(_3d_layers):
            sls.clear(layer)

        if renpy.config.missing_scene:
            renpy.config.missing_scene(layer)
    config.scene = scene
